<html>
<head>
    <title>Calliope results for model: {{ model_name }}</title>
    <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
    <style media="screen" type="text/css">
        html {
            font-family: Helvetica, sans-serif;
        }
        html, body {
            margin: 0;
            padding: 0;
        }
        #container {
            height: 100vh;
            width: 100vw;
        }
        #top {
            height: 50%;
            width: 100%;
        }
        #bottom_left {
            float: left;
            width: 50%;
            height: 50%;
        }
        #bottom_right {
            float: left;
            width: 50%;
            height: 50%;
        }
        #footer {
            clear: both;
        }
        h1 {
            font-size: 1.6em;
            font-weight: normal;
            border-bottom: 2px solid #CFCFCF;
            text-align: center;
        }
        h1 small {
            font-size: 0.6em;
        }
    </style>
</head>
<body>
    <div id="container">
        <h1>{{model_name}} <small>(run with Calliope {{ calliope_version }}, {{ result_stats }})</small></h1>
        <div id="top">{{ top }}</div>
        <div id="bottom_left">{{ bottom_left }}</div>
        <div id="bottom_right">{{ bottom_right }}</div>
        <div id="footer"></div>
    </div>
</body>
</html>
